/*
 Copyright (C) 2002 MandrakeSoft SA
 
 This program may be distributed and/or
 modified under the terms of the GNU General Public
 License version 2 as published by the Free Software Foundation
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.
 
 Licensees holding valid MandrakeSecurity MNF Commercial License may use this
 file in  accordance with the MandrakeSecurity MNF Commercial License Agreement
 provided with the Software.
 
 See http://www.mandrakesoft.com/mnf/license for information about MandrakeSecurity MNF Commercial License Agrrement.
*/ 
#include <stdlib.h>
#include "backend.h"

/* Compiler backend: store the instructions, and generate
   later the good ascii code; at this level resolve the
   goto labels */

static void print_goto(wvm_instruction_t *self, FILE *out);

/* utility, include single useless goto optimization */

static int get_target_pc(wvm_instruction_t *target)
{
  if (! target)
    {
      return -1;
    }
  else if (target->useless)
    return get_target_pc(target->next);
  else
    return target->pc;
}

/* Basic instruction emission */



static void print_push(wvm_instruction_t *self, FILE *out)
{
  fprintf(out, "%d\tPUSH\n", self->pc);
}


static void print_pop(wvm_instruction_t *self, FILE *out)
{
  fprintf(out, "%d\tPOP\n", self->pc);
}

static void print_halt(wvm_instruction_t *self, FILE *out)
{
  fprintf(out, "%d\tHALT\n", self->pc);
}


static void print_goto(wvm_instruction_t *self, FILE *out)
{
  /* Optmization: if the goto go the next instruction, do
     not generate the goto */
  
  wvm_goto_instruction_t *goto_self = (wvm_goto_instruction_t *) self;

  fprintf(out, "%d\tGOTO\t%d\n", self->pc, get_target_pc(goto_self->target));
}


static void print_exec(wvm_instruction_t *self, FILE *out)
{
  wvm_exec_instruction_t *exec_self = (wvm_exec_instruction_t *) self;

  fprintf(out, "%d\tEXEC\t%s\n", self->pc, exec_self->script);
}


static void print_if(wvm_instruction_t *self, FILE *out)
{
  wvm_if_instruction_t *if_self = (wvm_if_instruction_t *) self;

  fprintf(out, "%d\tIF\t%s\t%d\t%d\n", self->pc, if_self->status, get_target_pc(if_self->then_target), get_target_pc(if_self->else_target));
}

static const char  *wvm_page_prefix = "";

void wvm_set_page_prefix(const  char *page_prefix)
{
  wvm_page_prefix = page_prefix;
}

static void print_show(wvm_instruction_t *self, FILE *out)
{
  wvm_show_instruction_t *show_self = (wvm_show_instruction_t *) self;

  fprintf(out, "%d\tSHOW\t%s%s\n", self->pc, wvm_page_prefix, show_self->page);
}


#define wvm_init_instruction(instr, printer)   wvm_init_instruction_node((wvm_instruction_t *)(instr), (printer))
									 
static void wvm_init_instruction_node(wvm_instruction_t *instr, void (* printer)(wvm_instruction_t *self, FILE *out))
{
  /* Add it to the list */

  instr->next    = 0;
  instr->printer = printer;
  instr->pc      = 0;
  instr->useless = 0;
  instr->last =   instr;
}


wvm_instruction_t *wvm_make_push(void)
{
  wvm_instruction_t *instr;

  instr = (wvm_instruction_t *)malloc(sizeof(wvm_push_instruction_t));
  wvm_init_instruction(instr, print_push);
  
  return instr;
}

wvm_instruction_t *wvm_make_pop(void)
{
  wvm_instruction_t *instr;

  instr = (wvm_instruction_t *)malloc(sizeof(wvm_pop_instruction_t));
  wvm_init_instruction(instr, print_pop);
  
  return instr;
}


wvm_instruction_t *wvm_make_halt(void)
{
  wvm_instruction_t *instr;

  instr = (wvm_instruction_t *)malloc(sizeof(wvm_halt_instruction_t));
  wvm_init_instruction(instr, print_halt);
  
  return instr;
}


wvm_instruction_t *wvm_make_exec(const char *script)
{
  wvm_exec_instruction_t *instr;
  
  instr = (wvm_exec_instruction_t *)malloc(sizeof(wvm_exec_instruction_t));
  wvm_init_instruction(instr, print_exec);
  instr->script = script;
  
  return (wvm_instruction_t *) instr;
}


wvm_instruction_t *wvm_make_if(const char *status, wvm_instruction_t *then_target, wvm_instruction_t *else_target)
{
  wvm_if_instruction_t *instr;

  instr = (wvm_if_instruction_t *)malloc(sizeof(wvm_if_instruction_t));
  wvm_init_instruction(instr, print_if);
  instr->status = status;
  instr->then_target = then_target;
  instr->else_target = else_target;
  
  return (wvm_instruction_t *)instr;
}


wvm_instruction_t *wvm_make_goto(wvm_instruction_t *target)
{
  wvm_goto_instruction_t *instr;

  instr = (wvm_goto_instruction_t *)malloc(sizeof(wvm_goto_instruction_t));
  wvm_init_instruction(instr, print_goto);
  instr->target = target;
  
  return (wvm_instruction_t *)instr;
}


wvm_instruction_t *wvm_make_show(const char *page)
{
  wvm_show_instruction_t *instr;

  instr = (wvm_show_instruction_t *)malloc(sizeof(wvm_show_instruction_t));
  wvm_init_instruction(instr, print_show);
  instr->page = page;
  
  return (wvm_instruction_t *)instr;
}


/* Instruction list building */


void wvm_append(wvm_instruction_t *instr, wvm_instruction_t *append)
{
  instr->last->next = append;
  instr->last = append->last;
}

/* Optimizer */

static int useless_instruction(wvm_instruction_t *self)
{
  /* preserve generator optimizations */

  if (self->useless)
    return 1;

  if (self->printer == print_goto)
    {
      wvm_goto_instruction_t *goto_self = (wvm_goto_instruction_t *) self;

      return (goto_self->target == self->next);
    }
  else
    return 0;
}
  
/* Output code */

void wvm_print_code(wvm_instruction_t *instr, FILE *out)
{
  wvm_instruction_t *p;
  int count = 1;

  /* Count instructions, and optimize away useless instructions */

  for (p = instr; p; p = p->next)
    {
      if (! useless_instruction(p))
	p->pc = count++;
      else
	p->useless = 1;
    }

  for (p = instr; p; p = p->next)
    if (! p->useless)
      (* p->printer)(p, out);

}
