/*
 Copyright (C) 2002 MandrakeSoft SA
 
 This program may be distributed and/or
 modified under the terms of the GNU General Public
 License version 2 as published by the Free Software Foundation
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.
 
 Licensees holding valid MandrakeSecurity MNF Commercial License may use this
 file in  accordance with the MandrakeSecurity MNF Commercial License Agreement
 provided with the Software.
 
 See http://www.mandrakesoft.com/mnf/license for information about MandrakeSecurity MNF Commercial License Agrrement.
*/ 
#include "syntax.h"
#include "generator.h"



/* Utility: find out if a node code need accumator protection around */

static int need_protection(wat_node_t *node)
{
  /* Currently, no statement needs accumulator protection.
     For the moment, we leave the infrastructure in place,
     for future expansions. */

  return 0;
}


/* Generic node */

static wvm_instruction_t *container_generate(wat_node_t *node)
{
  wvm_instruction_t *instr = 0;
  wat_node_t *p;

  for (p = wat_node_get_first_child(node); p; p = wat_node_get_sibling(p))
    if (instr == 0)
      instr = wat_node_generate_code(p);
    else
      wvm_append(instr, wat_node_generate_code(p));

  return instr;
}


/* Generator for wizard */

static wvm_instruction_t *wizard_generate(wat_node_t *node)
{
  wvm_instruction_t *instr;

  /* Add a HALT as last instruction */

  instr = container_generate(node);
  wvm_append(instr, wvm_make_halt());

  return instr;
}


/* Generator for  If */

static wvm_instruction_t *if_generate(wat_node_t *node)
{
  wvm_instruction_t *first;
  wvm_instruction_t *exec;
  wvm_instruction_t *cond;
  wvm_instruction_t *then_body;
  wvm_instruction_t *else_body;
  wvm_instruction_t *last;
  
  /* Build first the instructions target for the if, then put them all together */

  first = wvm_make_push();
  exec  = wvm_make_exec(wat_if_node_get_test(node));
  last  = wvm_make_pop();
  then_body  = container_generate(wat_node_get_first_child_of_type(node, wat_then_node_type));

  {
    wat_node_t *p = wat_node_get_first_child_of_type(node, wat_else_node_type);

    if (p)
      else_body  = container_generate(p);
    else
      else_body  = 0;
  }

  /* Add to the then and else body the goto with the last as target */

  wvm_append(then_body, wvm_make_goto(last));
  
  if (else_body)
    wvm_append(else_body, wvm_make_goto(last));

  /* Make the if with the right targets */

  cond = wvm_make_if(wat_if_node_get_status(node), (then_body ? then_body : last), (else_body ? else_body : last));

  /* append all together on first */

  wvm_append(first, exec);
  wvm_append(first, cond);

  if (then_body)
    wvm_append(first, then_body);

  if (else_body)
    wvm_append(first, else_body);

  wvm_append(first, last);

  if (! need_protection(node))
    {
      first->useless = 1;
      last->useless  = 1;
    }

  return first;

}



/* Generator for Switch */


static wvm_instruction_t *case_default_generate(wat_node_t *node, wvm_instruction_t *last)
{
  wat_node_t *brother;
  wvm_instruction_t *code = 0;
  wvm_instruction_t *brother_code;

  brother = wat_node_get_sibling(node);

  if (brother)
    brother_code = case_default_generate(brother, last);
  else
    brother_code = 0;
  
  if ((! wat_node_is_empty(node)) && wat_node_is_case(node))
    {
      wvm_instruction_t *cond;
      wvm_instruction_t *body;

      body = container_generate(node);
      cond = wvm_make_if(wat_case_node_get_status(node), body, (brother_code ? brother_code : last));

      code = cond;
      wvm_append(code, body);

    }
  else if ((! wat_node_is_empty(node)) && wat_node_is_default(node))
    {
      code = container_generate(node);
    }

  if (code)
    {
      wvm_append(code, wvm_make_goto(last));

      if (brother_code)
	wvm_append(code, brother_code);

      return code;
    }
  else
    return brother_code;
}


static wvm_instruction_t *switch_generate(wat_node_t *node)
{
  wat_node_t *p;
  wvm_instruction_t *first;
  wvm_instruction_t *exec;
  wvm_instruction_t *body;
  wvm_instruction_t *last;

  /* Build first the instructions target for the if, then put them all together */

  first = wvm_make_push();
  exec  = wvm_make_exec(wat_switch_node_get_test(node));
  last  = wvm_make_pop();

  p  = wat_node_get_first_child_of_type(node, wat_case_node_type);

  if (! p) 
    p  = wat_node_get_first_child_of_type(node, wat_default_node_type);

  body  = case_default_generate(p, last);

  /* append all together on first */

  wvm_append(first, exec);

  if (body)
    wvm_append(first, body);

  wvm_append(first, last);

  return first;
}




/* Generator for While */


static wvm_instruction_t *while_generate(wat_node_t *node)
{
  wvm_instruction_t *first;
  wvm_instruction_t *exec;
  wvm_instruction_t *cond;
  wvm_instruction_t *body;
  wvm_instruction_t *last;
  
  /* Build first the instructions target for the if, then put them all together */

  first = wvm_make_push();
  exec  = wvm_make_exec(wat_while_node_get_test(node));
  last  = wvm_make_pop();
  body  = container_generate(node);

  /* Add to the body the goto with the exec as target */

  if (body)
    wvm_append(body, wvm_make_goto(exec));

  /* Make the if with the right targets */

  cond = wvm_make_if(wat_while_node_get_status(node), (body ? body : exec), last);

  /* append all together on first */

  wvm_append(first, exec);
  wvm_append(first, cond);

  if (body)
    wvm_append(first, body);

  wvm_append(first, last);

  if (! need_protection(node))
    {
      first->useless = 1;
      last->useless  = 1;
    }

  return first;
}


/* Generator for Page */

static wvm_instruction_t *page_generate(wat_node_t *node)
{
  return wvm_make_show(wat_page_node_get_name(node));
}


/* Generator for Call */

static wvm_instruction_t *call_generate(wat_node_t *node)
{
  wvm_instruction_t *instr;
  
  instr = wvm_make_push();
  wvm_append(instr, wvm_make_exec(wat_call_node_get_script(node)));
  wvm_append(instr, wvm_make_pop());

  return instr;
}



/* Types Descriptors */


static wat_node_type_descr_t wat_void_node_type_descr    = {0};
static wat_node_type_descr_t wat_root_node_type_descr    = {container_generate};
static wat_node_type_descr_t wat_wizard_node_type_descr  = {wizard_generate};
static wat_node_type_descr_t wat_if_node_type_descr      = {if_generate};
static wat_node_type_descr_t wat_then_node_type_descr    = {container_generate};
static wat_node_type_descr_t wat_else_node_type_descr    = {container_generate};
static wat_node_type_descr_t wat_switch_node_type_descr  = {switch_generate};
static wat_node_type_descr_t wat_case_node_type_descr    = {0};
static wat_node_type_descr_t wat_default_node_type_descr = {0};
static wat_node_type_descr_t wat_while_node_type_descr   = {while_generate};
static wat_node_type_descr_t wat_page_node_type_descr    = {page_generate};
static wat_node_type_descr_t wat_call_node_type_descr    = {call_generate};



/* Types IDs */

wat_node_type_t wat_void_node_type    = &wat_void_node_type_descr;
wat_node_type_t wat_root_node_type    = &wat_root_node_type_descr;
wat_node_type_t wat_wizard_node_type  = &wat_wizard_node_type_descr;
wat_node_type_t wat_if_node_type      = &wat_if_node_type_descr;
wat_node_type_t wat_then_node_type    = &wat_then_node_type_descr;
wat_node_type_t wat_else_node_type    = &wat_else_node_type_descr;
wat_node_type_t wat_switch_node_type  = &wat_switch_node_type_descr;
wat_node_type_t wat_case_node_type    = &wat_case_node_type_descr;
wat_node_type_t wat_default_node_type = &wat_default_node_type_descr;
wat_node_type_t wat_while_node_type   = &wat_while_node_type_descr;
wat_node_type_t wat_page_node_type    = &wat_page_node_type_descr;
wat_node_type_t wat_call_node_type    = &wat_call_node_type_descr;

/* Main generator */


void wat_generate_code(wat_node_t *node, FILE *out)
{
  wvm_instruction_t *code;

  code  = wat_node_generate_code(node);
  wvm_print_code(code, out);
}
