/*
 Copyright (C) 2002 MandrakeSoft SA
 
 This program may be distributed and/or
 modified under the terms of the GNU General Public
 License version 2 as published by the Free Software Foundation
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.
 
 Licensees holding valid MandrakeSecurity MNF Commercial License may use this
 file in  accordance with the MandrakeSecurity MNF Commercial License Agreement
 provided with the Software.
 
 See http://www.mandrakesoft.com/mnf/license for information about MandrakeSecurity MNF Commercial License Agrrement.
*/ 
#include <stdlib.h>
#include <string.h>
#include "parser.h"

/* Forward declarations for XML parsing functions */

static void parser_start_element(void *userData, const char *name, const char **atts);
static void parser_end_element(void *userData, const char *name);

static xmlSAXHandler wat_sax_parser = {
    0, /* internalSubset */
    0, /* isStandalone */
    0, /* hasInternalSubset */
    0, /* hasExternalSubset */
    0, /* resolveEntity */
    0, /* getEntity */
    0, /* entityDecl */
    0, /* notationDecl */
    0, /* attributeDecl */
    0, /* elementDecl */
    0, /* unparsedEntityDecl */
    0, /* setDocumentLocator */
    0, /* startDocument */
    0, /* endDocument */
    (startElementSAXFunc) parser_start_element, /* startElement */
    (endElementSAXFunc)   parser_end_element, /* endElement */
    0, /* reference */
    0,  /* characters */
    0, /* ignorableWhitespace */
    0, /* processingInstruction */
    0, /* comment */
    0, /* warning */
    0, /* error */
    0, /* fatalError */
    0, /* getParameterEntity */
    0 /*  cdataBlock */
};

/* ERROR CODES */

#define PARSING_OK         0
#define MISSING_ATTRIBUTE  1

typedef struct parsing_state
{
  wat_node_t *root;		/* The root node  */
  wat_node_t *current;		/* The current node being built, used during parsing */
  int         error;		/* The error code, if any */
  const char *info1;		/* additional info about the error */
  const char *info2;		/* additional info about the error */
} parsing_state_t;


static parsing_state_t *new_parsing_state(void)
{
  parsing_state_t *state;

  state = (parsing_state_t *) malloc(sizeof(parsing_state_t));
  state->root = 0;
  state->current = 0;
  state->error = PARSING_OK;
  state->info1 = 0;
  state->info2 = 0;

  return state;
}


static void delete_parsing_state(parsing_state_t *state)
{
  free(state);
}

static void set_error(parsing_state_t *state, int error, const char *info1, const char *info2)
{
  state->error   = error;
  state->info1 = info1;
  state->info2 = info2;
}


wat_node_t *wat_parse(const char *filename)
{
  parsing_state_t *state;
  xmlParserCtxtPtr ctxt;
  wat_node_t *root;


  ctxt = xmlCreateFileParserCtxt(filename);

  if (ctxt == 0)
    return 0;

  state = new_parsing_state();

  ctxt->sax = &wat_sax_parser;
  ctxt->userData = (void *)state;

  xmlParseDocument(ctxt);

  ctxt->sax = NULL;
  xmlFreeParserCtxt(ctxt);

  if (state->error == PARSING_OK)
    {
      root = state->root;
      delete_parsing_state(state);

      return root;
    }
  else
    {
      switch (state->error)
	{
	case MISSING_ATTRIBUTE:
	  fprintf(stderr, "Missing attribute %s for node of tag %s\n", state->info1, state->info2);
	  break;
	default:
	  fprintf(stderr, "Unknown parsing error: %s  %s\n", state->info1, state->info2);
	  break;
	}

      delete_parsing_state(state);
      return 0;
    }
}


 // Add the node as children of current, move current to the node.

static void push_node(parsing_state_t *state, wat_node_t *n)
{
  if (state->current == 0)
    {
      state->root =  n;
      state->current =  n;
    }
  else
    {
      wat_node_add_child(state->current, n);
      state->current = n;
    }
}


/** Complete the current node, and make its parent the current node */


static void pop_node(parsing_state_t *state)
{
  /* Pop to the current node parent */
  
  state->current = wat_node_get_parent(state->current);
}


/* Parsing functions */

static const char *get_attribute(parsing_state_t *state, const char **atts, const char *name, const char *tag)
{
  const char *value = 0;
  int i;

  if (atts)
    for (i = 0; atts[i] != 0; i += 2)
      if (! strcmp(name,  atts[i]))
	value = atts[i+1];

  if (value == 0)
    {
      value = "#unknown#";
      set_error(state, MISSING_ATTRIBUTE, name, tag);
    }

  return value;
}


static void parser_start_element(void *userData, const char *name, const char **atts)
{
  parsing_state_t *state = (parsing_state_t *) userData;
  wat_node_t *n;

  /* Built the new element */

  if (! strcmp(name, WAT_TAG))
    {
      n = wat_node_new_root();
      push_node(state, n);
    }
  else if (! strcmp(name, WIZARD_TAG))
    {
      n = wat_node_new_wizard();
      push_node(state, n);
    }
  else if (! strcmp(name, PAGE_TAG))
    {
      n = wat_node_new_page(get_attribute(state, atts, "name", PAGE_TAG));
      push_node(state, n);
    }
  else if (! strcmp(name, IF_TAG))
    {
      n = wat_node_new_if(get_attribute(state, atts, "test", IF_TAG), get_attribute(state, atts, "status", IF_TAG));
      push_node(state, n);
    }
  else if (! strcmp(name, THEN_TAG))
    {
      n = wat_node_new_then();
      push_node(state, n);
    }
  else if (! strcmp(name, ELSE_TAG))
    {
      n = wat_node_new_else();
      push_node(state, n);
    }
  else if (! strcmp(name, SWITCH_TAG))
    {
      n = wat_node_new_switch(get_attribute(state, atts, "test", SWITCH_TAG));
      push_node(state, n);
    }
  else if (! strcmp(name, CASE_TAG))
    {
      n = wat_node_new_case(get_attribute(state, atts, "status", CASE_TAG));
      push_node(state, n);
    }
  else if (! strcmp(name, DEFAULT_TAG))
    {
      n = wat_node_new_default();
      push_node(state, n);
    }
  else if (! strcmp(name, WHILE_TAG))
    {
      n = wat_node_new_while(get_attribute(state, atts, "test", WHILE_TAG), get_attribute(state, atts, "status", WHILE_TAG));
      push_node(state, n);
    }
  else if (! strcmp(name, CALL_TAG))
    {
      n = wat_node_new_call(get_attribute(state, atts, "script", CALL_TAG));
      push_node(state, n);
    }
  else
    {
      /* Ignore */
    }
}


static void parser_end_element(void *userData, const char *name)
{
  parsing_state_t *state = (parsing_state_t *) userData;

  if (! strcmp(name, WAT_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, WIZARD_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, PAGE_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, IF_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, THEN_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, ELSE_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, SWITCH_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, CASE_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, DEFAULT_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, WHILE_TAG))
    {
      pop_node(state);
    }
  else if (! strcmp(name, CALL_TAG))
    {
      pop_node(state);
    }
  else
    {
      /* Ignore */
    }
}









