
package com.mandrakesoft.wat;

import org.apache.xerces.parsers.DOMParser;
import org.apache.xerces.dom.NodeImpl;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import java.io.IOException;
import java.io.*;
import java.util.*;

/**
 * Compiler for the Wat xml files. Uses DOM to parse the input and the DOM tree as syntactic tree. 
 * The compilation is a three-step process:
 * 1) Mark all the syntactic  nodes with a "line number"
 * 2) Static analysis: determines the next node of every syntactic nodes
 * 3) Code generation.
 *
 * The Code generation is done by a Generator object, contained in a Property file. 
 * This makes simpler to substitute the front end of the compiler without modifying its code.
 * Currently the syntactic tree is a view of the DOM tree filtered out by a set of functions 
 * (isSyntacticNode, parent, sibling). 
 * This view should probably be implemented with some sort of TreeWalker/NodeFilter instead.
 */
public class Compiler {
  
  static int global_count = 1;
  Generator generator;

  /* Constructor. Parses the wat file, perform a static flow analysis, and generates the code
   * using the given front-end */

  public Compiler(InputSource xmlFile, String codeGenerator) 
  {
    DOMParser parser = new DOMParser();
    
    try
      {
	parser.parse(xmlFile);
      }
    catch (SAXException se)
      {
	se.printStackTrace();
      }
    catch (IOException ioe)
      {
	ioe.printStackTrace();
      }
    
    Document document = parser.getDocument();

    document.getDocumentElement().normalize();

    try
      {
	generator = (Generator) Class.forName(codeGenerator).newInstance();
      }
    catch(Exception e)
      {
	System.err.println();

	System.err.println("Can't mount the compiler front-end: " + codeGenerator);
	System.err.println(e);
	System.exit(0);
      }

    compile(document);
  }


  /**
   * The main code generation routine. It numbers the nodes, calculates the flow, and generate the output. */

  public void compile(Document doc)
  {
    Node node = (Node) doc.getDocumentElement();
    numberNodes(node);
    calculateNext(node);
    generator.generateCode(doc);
  }


  /**
   * Associate a number (something like a line number) to each node of the syntactic tree
   * Just assign an increasing number to nodes correspoding to an executable state and do the same for
   * the children */

  private void numberNodes(Node node)
  {
    if (Syntax.isSyntacticNode(node))
      {
	if (Syntax.isExecutable(node))
	  Syntax.setCompilerInfo(node, new CompilerInfo(global_count++));

	for (int i=0; i<node.getChildNodes().getLength(); i++)
	  {
	    numberNodes((Node) node.getChildNodes().item(i));
	  }
      }
  }


  /**
   * Foreach node in the tree, find the next node in the (static) flow.
   * If the node is the "Wizard" node, the next is NONE.
   * If the node's next sibling 1) exists the next node is the sibling.
   * Other wise, next node is the sibling of the first executable ancestor.
   */

  private void calculateNext(Node node)
  {
    String tagName;
    CompilerInfo nodeInfo = Syntax.getCompilerInfo(node);

    if (Syntax.isExecutable(node))
      {
	if (Syntax.getSibling(node) != null)
	  {
	    CompilerInfo nextInfo = Syntax.getCompilerInfo(Syntax.getSibling(node));
	    nodeInfo.setNext(nextInfo.getLineNo());
	  }
	else
	  {
	    Node ancestor = Syntax.getParent(node);

	    while ((ancestor != null) && (! Syntax.isExecutable(ancestor)))
	      ancestor = Syntax.getParent(ancestor);

	    if (ancestor == null)
	      nodeInfo.setNext(CompilerInfo.NONE);
	    else if (Syntax.isALoop(ancestor))
	      nodeInfo.setNext(Syntax.getCompilerInfo(ancestor).getLineNo());
	    else
	      nodeInfo.setNext(Syntax.getCompilerInfo(ancestor).getNext());
	  }
      }
    
    
    for (int i=0; i<node.getChildNodes().getLength(); i++)
      {
	if (Syntax.isSyntacticNode(node))
	    calculateNext((Node) node.getChildNodes().item(i));
      }
  }
  
  
  private static void usage()
  {
    System.err.println("usage: Compiler <file.wat> <property file>");
  }

  /**
   * Her majesty.
   * It accepts a path to a wat file to compile and to a property file containing the front-end specification, 
   * checks the arguments, build an InputSource and a Compiler instance with it.
   */

  public static void main( String argv[])
  {
    FileInputStream file = null;

    if (argv.length != 2) { usage(); System.exit(0); }

    try {
      file = new FileInputStream(argv[0]);
    }
    catch (FileNotFoundException e)
      {
	System.err.println("Input file " + argv[0] + "not found");
	System.exit(0);
      }

    Properties p = new Properties();

    try {
      p.load(new FileInputStream(argv[1])); 
    }
    catch (IOException e)
      {
	System.err.println("Can't load property file: " + argv[1]);
	System.exit(0);
      }

    new Compiler(new InputSource(file), p.getProperty("front-end"));
  }
}
