
package com.mandrakesoft.wat;

import org.w3c.dom.*;
import org.apache.xerces.dom.NodeImpl;

/**
 * A class that offers a set of static methods to filter out syntactic nodes from
 * other nodes (ex. text nodes, or format description)
 * and to classify nodes from the syntactic point of view.
 */

public class Syntax
{
  public static final String watTag    = "Wat";
  public static final String wizardTag = "Wizard";
  public static final String pageTag   = "Page";
  public static final String ifTag     = "If";
  public static final String thenTag   = "Then";
  public static final String elseTag   = "Else";
  public static final String whileTag  = "While";

  /**
   * Find the syntactic parent of the node: the first ancestor that is a syntactic node */

  public static Node getParent( Node node)
  {
    if (node == null || node.getParentNode() == null )
      return null;
    
    if (! isSyntacticNode(node.getParentNode()))
      return getParent(node.getParentNode());
    else
      return node.getParentNode();
  }

  /**
   * Find the syntactic sibling of the node: the first sibling that is a syntactic node. */

  public static Node getSibling(Node node)
  {
    if (node == null)
      return null;
    else if (isSyntacticNode(node.getNextSibling()))
      return node.getNextSibling();
    else
      return getSibling(node.getNextSibling());
  }

  /**
   * Find the first child of a node (that is syntactic node), if any  */

  public static Node getFirstChild(Node node)
  {
    for (int i=0; i < node.getChildNodes().getLength(); i++)
      {
	Node child = (Node) node.getChildNodes().item(i);
	
	if (isSyntacticNode(child))
	  return child;
      }

    return null;
  }

  /**
   * Find the first element child of a node  with a given tag, if any */

  public static Node getFirstChild(Node node, String tag)
  {
    for (int i=0; i < node.getChildNodes().getLength(); i++)
      {
	Node child = (Node) node.getChildNodes().item(i);
	
	if ((child instanceof Element) && ((Element)child).getTagName().equals(tag))
	  return child;
      }

    return null;
  }

  /**
   * Get a node attribute */

  public static String getAttribute(Node node, String attribute)
  {
    if (! (node instanceof Element))
      return null;
    else
      return ((Element)node).getAttribute(attribute);
  }

  /**
   * A node that have a syntactic role (not text nodes and so...) */

  public static boolean isSyntacticNode(Node node)
  {
    if (! (node instanceof Element))
      return false;
    else
      {
	String tag = ((Element)node).getTagName();

	return (tag.equals(watTag)    ||
		tag.equals(wizardTag) ||
		tag.equals(pageTag)   ||
		tag.equals(ifTag)     ||
		tag.equals(thenTag)   ||
		tag.equals(elseTag)   ||
		tag.equals(whileTag));
      }
  }



  public static boolean isRoot(Node node)
  {
    return node.getParentNode() == null;
  }

  public static boolean isWat(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(watTag);
  }

  public static boolean isWizard(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(wizardTag);
  }

  public static boolean isPage(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(pageTag);
  }

  public static boolean isWhile(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(whileTag);
  }

  public static boolean isIf(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(ifTag);
  }

  public static boolean isThen(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(thenTag);
  }

  public static boolean isElse(Node node)
  {
    return (node instanceof Element) && (((Element)node).getTagName()).equals(elseTag);
  }

  /** Check if a node is executable. An  executable node is a node that have
      an associated "vm machine state" in  the generated code  */

  public static boolean isExecutable(Node node)
  {
    if (! (node instanceof Element))
      return false;
    else
      {
	String tag = ((Element)node).getTagName();

	return (tag.equals(pageTag) ||
		tag.equals(ifTag) ||
		tag.equals(whileTag));
      }
  }

  /** Check if a node is a loop, i.e. if the last node next is the top level node. */

  public static boolean isALoop(Node node)
  {
    return  isWhile(node);
  }

  /** Utility function. Returns the UserData of the node as a compilerinfo object. */

  public static CompilerInfo getCompilerInfo(Node node)
  {
    return (CompilerInfo) (((NodeImpl)node).getUserData());
  }

  /** Utility function. Set the UserData of the node, as a compilerinfo object. */

  public static void setCompilerInfo(Node node, CompilerInfo info)
  {
    ((NodeImpl)node).setUserData(info);
  }
}
